import 'package:flutter_secure_storage/flutter_secure_storage.dart';
import 'package:senior_platform_authentication/senior_platform_authentication.dart';

import '../../domain/entities/secure_storage_keys.dart';

abstract class SecureStorageDatasource {
  Future<void> deleteUser();
  Future<void> deleteToken();
  Future<void> deleteTokenExpirationDate();
  Future<UserModel?> readUser();
  Future<TokenModel?> readToken();
  Future<DateTime?> readTokenExpirationDate();
  Future<void> writeUser(UserModel user);
  Future<void> writeToken(TokenModel token);
  Future<void> writeTokenExpirationDate(DateTime tokenExpirationDate);
}

class SecureStorageLocalDatasource implements SecureStorageDatasource {
  late final FlutterSecureStorage _storage;

  SecureStorageLocalDatasource({FlutterSecureStorage? storage}) {
    _storage = storage ?? const FlutterSecureStorage();
  }

  @override
  Future<UserModel?> readUser() async {
    final jsonStr = await _storage.read(key: SecureStorageKeys.user);

    if (jsonStr != null) {
      return UserModel.fromJson(jsonStr);
    }

    return null;
  }

  @override
  Future<TokenModel?> readToken() async {
    final jsonStr = await _storage.read(key: SecureStorageKeys.token);

    if (jsonStr != null) {
      return TokenModel.fromJson(jsonStr);
    }

    return null;
  }

  @override
  Future<DateTime?> readTokenExpirationDate() async {
    final jsonStr =
        await _storage.read(key: SecureStorageKeys.tokenExpirationDate);

    if (jsonStr != null) {
      return DateTime.tryParse(jsonStr);
    }

    return null;
  }

  @override
  Future<void> writeUser(UserModel user) async {
    await _storage.write(key: SecureStorageKeys.user, value: user.toJson());
  }

  @override
  Future<void> writeToken(TokenModel token) async {
    await _storage.write(key: SecureStorageKeys.token, value: token.toJson());
  }

  @override
  Future<void> writeTokenExpirationDate(DateTime tokenExpirationDate) async {
    await _storage.write(
        key: SecureStorageKeys.tokenExpirationDate,
        value: tokenExpirationDate.toIso8601String());
  }

  @override
  Future<void> deleteUser() async {
    await _storage.delete(key: SecureStorageKeys.user);
  }

  @override
  Future<void> deleteToken() async {
    await _storage.delete(key: SecureStorageKeys.token);
  }

  @override
  Future<void> deleteTokenExpirationDate() async {
    await _storage.delete(key: SecureStorageKeys.tokenExpirationDate);
  }
}
