import 'dart:developer';

import 'package:equatable/equatable.dart';
import 'package:flutter_bloc/flutter_bloc.dart';

import '../../../../../senior_platform_authentication_ui.dart';

part 'authentication_event.dart';
part 'authentication_state.dart';

class AuthenticationBloc
    extends Bloc<AuthenticationEvent, AuthenticationState> {
  final GetUserUsecase _getUserUsecase;
  final StoreAuthenticationDataUsecase _storeAuthenticationDataUsecase;
  final CheckStoredTokenUsecase _checkStoredTokenUsecase;
  final LogoutUsecase _logoutUsecase;
  final ClearStoredDataUsecase _clearStoredDataUsecase;
  final GetStoredTokenUsecase _getStoredTokenUsecase;
  final RefreshStoredTokenUsecase _refreshStoredTokenUsecase;

  /// Internal constructor, should not be exposed.
  AuthenticationBloc._internal({
    required GetUserUsecase getUserUsecase,
    required StoreAuthenticationDataUsecase storeAuthenticationDataUsecase,
    required CheckStoredTokenUsecase checkStoredTokenUsecase,
    required LogoutUsecase logoutUsecase,
    required ClearStoredDataUsecase clearStoredDataUsecase,
    required GetStoredTokenUsecase getStoredTokenUsecase,
    required RefreshStoredTokenUsecase refreshStoredTokenUsecase,
  })  : _getUserUsecase = getUserUsecase,
        _storeAuthenticationDataUsecase = storeAuthenticationDataUsecase,
        _checkStoredTokenUsecase = checkStoredTokenUsecase,
        _logoutUsecase = logoutUsecase,
        _clearStoredDataUsecase = clearStoredDataUsecase,
        _getStoredTokenUsecase = getStoredTokenUsecase,
        _refreshStoredTokenUsecase = refreshStoredTokenUsecase,
        super(const AuthenticationState.unknown()) {
    on<AuthenticationStatusChanged>(_onAuthenticationStatusChanged);
    on<LogoutRequested>(_onLogoutRequested);
    on<CheckAuthenticationRequested>(_onCheckAuthenticationRequested);
  }

  /// Singleton instance for production environment.
  static final AuthenticationBloc _instance = AuthenticationBloc._internal(
    getUserUsecase: GetUserUsecase(),
    storeAuthenticationDataUsecase: StoreAuthenticationDataUsecase(),
    checkStoredTokenUsecase:
        CheckStoredTokenUsecase(getUserUsecase: GetUserUsecase()),
    logoutUsecase: LogoutUsecase(),
    clearStoredDataUsecase: ClearStoredDataUsecase(),
    getStoredTokenUsecase: GetStoredTokenUsecase(),
    refreshStoredTokenUsecase: RefreshStoredTokenUsecase(
      getUserUsecase: GetUserUsecase(),
      refreshTokenUsecase: RefreshTokenUsecase(),
      storeAuthenticationDataUsecase: StoreAuthenticationDataUsecase(),
      clearStoredDataUsecase: ClearStoredDataUsecase(),
    ),
  );

  /// Factory exposed constructor.
  factory AuthenticationBloc() => _instance;

  /// Factory mocked exposed constructor. Should only be used for tests.
  factory AuthenticationBloc.mocked({
    required GetUserUsecase getUserUsecase,
    required StoreAuthenticationDataUsecase storeAuthenticationDataUsecase,
    required CheckStoredTokenUsecase checkStoredTokenUsecase,
    required LogoutUsecase logoutUsecase,
    required ClearStoredDataUsecase clearStoredDataUsecase,
    required GetStoredTokenUsecase getStoredTokenUsecase,
    required RefreshStoredTokenUsecase refreshStoredTokenUsecase,
  }) {
    return AuthenticationBloc._internal(
      getUserUsecase: getUserUsecase,
      storeAuthenticationDataUsecase: storeAuthenticationDataUsecase,
      checkStoredTokenUsecase: checkStoredTokenUsecase,
      logoutUsecase: logoutUsecase,
      clearStoredDataUsecase: clearStoredDataUsecase,
      getStoredTokenUsecase: getStoredTokenUsecase,
      refreshStoredTokenUsecase: refreshStoredTokenUsecase,
    );
  }

  Future<void> _onAuthenticationStatusChanged(
    AuthenticationStatusChanged event,
    Emitter<AuthenticationState> emit,
  ) async {
    switch (event.status) {
      case AuthenticationStatus.unauthenticated:
        return emit(const AuthenticationState.unauthenticated());
      case AuthenticationStatus.authenticated:
        final user = await _tryGetUser(
            event.authenticationResponse?.token?.accessToken ?? '');
        final stored = await _storeAuthenticationDataUsecase(
          AuthenticationData(
            user: user,
            token: event.authenticationResponse?.token,
          ),
        );
        return emit(stored
            ? const AuthenticationState.authenticated()
            : const AuthenticationState.unauthenticated());
      case AuthenticationStatus.unknown:
        return emit(const AuthenticationState.unknown());
    }
  }

  void _onLogoutRequested(
    LogoutRequested event,
    Emitter<AuthenticationState> emit,
  ) async {
    try {
      final token = await _getStoredTokenUsecase(NoParams());
      if (token != null) {
        await _logoutUsecase(token.accessToken);
      }
    } catch (err) {
      log(err.toString());
    } finally {
      await _clearStoredDataUsecase(NoParams());
      emit(const AuthenticationState.unauthenticated());
    }
  }

  void _onCheckAuthenticationRequested(
    CheckAuthenticationRequested event,
    Emitter<AuthenticationState> emit,
  ) async {
    final isTokenValid = await _checkStoredTokenUsecase(NoParams());

    if (!isTokenValid) {
      final token = await _refreshStoredTokenUsecase(NoParams());

      return emit(token != null
          ? const AuthenticationState.authenticated()
          : const AuthenticationState.unauthenticated());
    }

    return emit(const AuthenticationState.authenticated());
  }

  Future<User?> _tryGetUser(String accessToken) async {
    try {
      final user = await _getUserUsecase(accessToken);
      return user;
    } catch (_) {
      return null;
    }
  }
}
